<?php

require_once("Control/GameConstants.inc");
require_once("Control/Utils.inc");
require_once("Control/INavigable.inc");
require_once("Control/ShipActionController.inc");
require_once("Model/SolarSystem.inc");
require_once("Model/Planet.inc");
require_once("Model/Ship.inc");

/**
 * Controls flight activities, navigates ships.
 * @author Petr Kellnhofer (A07278)
 *
 */
class FlightController
{
	const PATH_CALC_DT = 86400; // sec
	
	/**
	 * Solar system.
	 * @var SolarSystem
	 */
	private $solarSystem;
	
	private $shipActionController;
	
	private $ship;
	
	/**
	 * Creates new instance of controller.
	 * @param SolarSystem $solarSystem associated solarSystem
	 * @return new instance
	 */
	public function __construct(ShipActionController $shipActionController)
	{
		$this->shipActionController = $shipActionController;
		$this->solarSystem = $shipActionController->GetSolarSystem();
		$this->ship = $shipActionController->GetShip();
	}
	
	/**
	 * Detects planet on which can specified ship dock according to current position.
	 * @return near planet or NULL 
	 */
	public function GetDockablePlanet(array $position = NULL)
	{
		$maxDockDistanceSq = GameConstants::MAX_DOCK_DISTANCE * GameConstants::MAX_DOCK_DISTANCE;
		
		$planets = (array)$this->solarSystem->GetPlanets();
		
		foreach ($planets as $planet)
		{
			if ($filghter->GetDistanceToPlanet($planet, $position) <= GameConstants::MAX_DOCK_DISTANCE)
			{
				return $planet;
			}
		}
		
		return NULL;
	}
	
	/**
	 * Calculates distance between planets.
	 * @param Planet $planet1 first planet
	 * @param Planet $planet2 second planet
	 * @return distance between planets in game units
	 */
	public static function GetPlanetDistance(Planet $planet1, Planet $planet2)
	{
		$dx = $planet1->GetWorldX() - $planet2->GetWorldX();
		$dy = $planet1->GetWorldY() - $planet2->GetWorldY();
		
		return sqrt($dx * $dx + $dy * $dy);
	}
	
	/**
	 * Gets distance of ship to specified planet.
	 * @param Planet $planet planet to measure distance to
	 * @return distance of ship to specified planet
	 */
	public function GetDistanceToPlanet(Planet $planet, array $position = NULL)
	{
		if ($position == NULL)
		{
			$position = array();
			$position[0] = $this->ship->GetX();
			$position[1] = $this->ship->GetY();
		}
		
		$dx = $planet->GetWorldX() - $position[0];
		$dy = $planet->GetWorldY() - $position[1];
		
		return sqrt($dx * $dx + $dy * $dy);
	}
	
	private function InterpolatePositions(array $posA, array $posB, $ratio)
	{
		$res = array();
		$res[0] = $posA[0] + ($posB[0] - $posA[0]) * $ratio;
		$res[1] = $posA[1] + ($posB[1] - $posA[1]) * $ratio;
		return $res;
	}
	
	/**
	 * 
	 * @param $planet
	 * @param $startGameDay
	 * @param $elapsedGameTimeSec
	 * @return array[0] = true if planet reached
	 * 			array[1] = path as array of position arrays
	 * 			array[2] = time spent by travelling (until reaching planet or $elapsedGameTimeSec spent)
	 * 			array[3] = number of direction changes
	 */
	public function FlyToTarget(INavigable $navigableObject, $startGameDayFloat, $elapsedGameTimeSec)
	{		
		$startRemTime = $elapsedGameTimeSec;
		
		$res = array();
		$res[0] = false;
		$path = array();
		$dirChangeCount = 0;
		
		$speed = $this->ship->GetModel()->GetSpeed();
		$position = array();
		$position[0] = $this->ship->GetX();
		$position[1] = $this->ship->GetY();
		
		while ($elapsedGameTimeSec > 0)
		{
			$prevPos = $position;
			
			$time = min($elapsedGameTimeSec, FlightController::PATH_CALC_DT);
			$distance = $speed * $time;
			
			$posA = $navigableObject->GetPositionForGameDay(floor($startGameDayFloat), $this->solarSystem->GetId());
			$posB = $navigableObject->GetPositionForGameDay(floor($startGameDayFloat) + 1, $this->solarSystem->GetId());
			$target = $this->InterpolatePositions($posA, $posB, $startGameDayFloat - floor($startGameDayFloat));
			
			$direction = Utils::SubtractVertices2($target, $position);
			$dirChangeCount++;
			
			$dirSize = Utils::GetVector2Length($direction);
			//echo "can do: " . (($distance + GameConstants::MAX_DOCK_DISTANCE)/1000000.0) . "<br/>";
			//echo "must do: " . ($dirSize/1000000.0) . "<br/>"; 
			
			if ($dirSize <= $distance + GameConstants::MAX_DOCK_DISTANCE)
			{
				array_push($path, array(0 => $posA[0], 1 => $posB[1]));
				$res[0] = true;	
				break;
			}
			else
			{
				$direction = Utils::MultiplyVertex2($direction, $distance / floatval($dirSize));
				$position = Utils::AddVertices2($position, $direction);		
				array_push($path, array(0 => $position[0], 1 => $position[1]));
			}		
			
			$startGameDayFloat += $time / floatval(GameTimer::SECONDS_IN_DAY);
			$elapsedGameTimeSec -= $time;
		}
		
		$res[1] = $path;
		$res[2] = $startRemTime - $elapsedGameTimeSec;
		$res[3] = $dirChangeCount;
		
		return $res;
	}
}

?>