<?php

require_once("Model/SolarSystem.inc");
require_once("Model/Planet.inc");
require_once("Model/GoodType.inc");
require_once("Model/Player.inc");
require_once("Model/Ship.inc");
require_once("Control/GameConstants.inc");
require_once("Control/Utils.inc");
require_once("Control/Responder.inc");
require_once("Control/FlightController.inc");
require_once("Control/ActionLogger.inc");
require_once("Control/ShipActionController.inc");

/**
 * Updates DB state, creates deliveries, plays for PC player.
 */
class GameUpdater
{
	/**
	 * Loaded solar system.
	 */
	private $solarSystem;
	
	/**
	 * List of available goods or null.
	 */
	private $goods;
	
	const NPC_CARGO_SHIP_MODEL_ID = 1;
	const NPC_GUARD_SHIP_MODEL_ID = 4;
	
	/**
	 * Creates new updater for specified solar system.
	 * @param solarSystem solar system
	 */
	public function __construct(SolarSystem $solarSystem)
	{
		if ($solarSystem == NULL)
		{
			throw new Exception("Solar system must not be null."); 
		}
		
		$this->solarSystem = $solarSystem;
		$this->goods = NULL;
	}
	
	public function UpdateAll()
	{
		$this->PlayPCPlayer();
		$this->CheckDeliveries();
		$this->UpdatePirates();
	} 
	
	/**
	 * Checks deliveries on planet. Generates new if needed.
	 */
	public function CheckDeliveries()
	{
		foreach ($this->solarSystem->GetPlanets() as $planet)
		{
			$this->CheckDeliveriesOnPlanet($planet);
		}
	}	
	
	public function CheckDeliveriesOnPlanet(Planet $planet)
	{
		$deliveries = $planet->GetDeliveries();
		if ($deliveries == NULL)
		{
			$planet->LoadDeliveriesFromDB();
			$deliveries = $planet->GetDeliveries();				
		}
		
		if (count($deliveries) < GameConstants::OPTIMAL_DELIVERIES_PER_PLANET)
		{
			$this->GenerateDeliveriesForPlanet($planet, GameConstants::OPTIMAL_DELIVERIES_PER_PLANET - count($deliveries));
		}
	}
	
	public function UpdatePirates()
	{
		echo "<h3>Hra za piraty</h3>";
		$pirates = Player::LoadPiratesFromDB();
		
		if ($pirates->GetLastAction() + GameConstants::PIRATE_UPDATE_PERIOD_REAL_SEC > time())
		{
			echo "<i>nebudu hrat za piraty - jeste nenastal cas</i><br/>";
			return;
		}
		
		$this->ReducePirateRelations($pirates);
		$this->AttackEnemyFleets($pirates);		
		$pirates->MarkLastAction();	
	}
	
	private function AttackEnemyFleets(Player $pirates)
	{
		echo "<h4>Utoky piratu</h4>";
		$pirateShips = $pirates->GetShipsForced();
		
		$players = (array)Player::LoadAllFromDB();
		$fleets = array();
		
		foreach ($players as $player)
		{
			if ($player->IsPirate() || Utils::GetPlayersAlliancePirateRelations($player) > 0)
			{
				continue;
			}
			
			$ships = $player->GetShipsForced();
			
			foreach ($ships as $ship)
			{
				if ($ship->GetModel()->GetShipType()->GetId() != GameConstants::CARGO_SHIP_TYPE_ID)
				{
					continue;
				}
				
				$shipAC = new ShipActionController($player, $ship);
				$shipAC->UpdateState();
				if ($shipAC->IsFlying())
				{
					array_push($fleets, $ship);
				}
			}
		}
		
		foreach ($fleets as $cargo)
		{
			$dictionary = Dictionary::GetInstance();
			$luck = Utils::GetRandomProbability();
			if ($luck > GameConstants::PIRATE_ATTACK_FLEET_PROBABILITY)
			{
				printf("Lod %s hrace ma stesti.<br/>", $cargo->GetName());
				continue;
			}
			
			$player = Player::LoadFromDB($cargo->GetPlayerID());	
			$guards = Utils::GetAssignedGuardians($cargo, $player);
			$guardCount = count($guards);
			
			$pirateTotalCount = count($pirateShips);
			
			if ($pirateTotalCount <= 0)
			{
				printf("Piratum dosly lode.<br/>");
				return;
			}
			
			$pirateCount = min($pirateTotalCount, floor($guardCount * GameConstants::PIRATE_ATTACK_ADVANTAGE_RATIO));
			
			printf("Lod %s hrace %s s %d obranci bude napadena %d piraty.<br/>", $cargo->GetName(), $player->GetStringIdentifier(), $guardCount, $pirateCount);
			$ourLoss = 0;
			$pirateLoss = 0;
			
			while (true)
			{
				// pirate fire
				if (Utils::GetRandomProbability() > 0.5)
				{
					$guard = array_pop($guards);
					printf("Obrance %s hrace %s byl znicen.<br/>", $guard->GetName(), $player->GetStringIdentifier());
					$guard->RemoveFromDB();
					
					$guardCount--;
					$ourLoss++;
					if ($guardCount <= 0)
					{
						break;
					}
				}
				
				// player fire
				if (Utils::GetRandomProbability() > 0.5)
				{
					$pirateShip = array_pop($pirateShips);
					printf("Utocnik %s piratu byl znicen.<br/>", $pirateShip->GetName());
					$pirateShip->RemoveFromDB();					

					$pirateCount--;
					$pirateLoss++;
					if ($pirateCount <= 0)
					{
						break;
					}
				}
			}
			
			// player won
			if ($guardCount > 0)
			{
				printf("Hrac %s vyhral.<br/>", $player->GetStringIdentifier());
				$logger = new ActionLogger($player);
				$logger->LogPiratesAttackWon($cargo, $ourLoss, $pirateLoss);
				
				$player->SetPirateRelations($player->GetPirateRelations() - 1);
				$logger->LogPiratesRelationsDecreased();
			}
			else
			{
				// pirates won
				printf("Hrac %s prohral.<br/>", $player->GetStringIdentifier());

				$cargo->SetLevel($cargo->GetLevel() - $pirateCount);
				
				$logger = new ActionLogger($player);
				$logger->LogPiratesAttackLost($cargo, $ourLoss, $pirateLoss, $pirateCount);
				
				if ($cargo->GetLevel() <= 0)
				{
					$logger->LogPiratesCargoLost($cargo);
					$cargo->RemoveFromDB();
				}				
			}			
		}
	}
	
	private function ReducePirateRelations(Player $pirates)
	{
		echo "<h4>Redukce vztahu s piraty</h4>";
		$players = (array)Player::LoadAllFromDB();
		$friends = array();
		
		foreach ($players as $player)
		{
			if ($player->IsPCPlayer() || $player->IsPirate())
			{
				continue;
			}
			
			if (Utils::GetPlayersAlliancePirateRelations($player) > 0)
			{
				array_push($friends, $player);
			}
		}
		
		$friendCount = count($friends);
		$reduceCount = $friendCount * GameConstants::PIRATE_RELATION_REDUCTION_PLAYER_RATIO; 
		shuffle($friends);
		
		for ($i = 0; $i < $reduceCount; $i++)
		{
			$player = array_pop($friends);
			$player->SetPirateRelations($player->GetPirateRelations() - 1);
			$logger = new ActionLogger($player);
			$logger->LogPiratesRelationsDecreased();
			printf("pirati ted kamaradi s %s na %d<br/>", $player->GetStringIdentifier(), Utils::GetPlayersAlliancePirateRelations($player));
			
			if (Utils::GetPlayersAlliancePirateRelations($player) <= 0)
			{
				if ($player->GetAllianceId() != NULL)
				{
					$members = (array)Alliance::LoadFromDB($player->GetAllianceId())->GetMembersForced();	
				}
				else
				{
					$members = array(0 => $player);
				}
				
				foreach ($members as $member)
				{
					$this->NotifyPlayerNotFriendAnymore($member, $pirates);
				}				
			}
		}
	}
	
	private function NotifyPlayerNotFriendAnymore(Player $player, Player $pirates)
	{
		$dictionary = Dictionary::GetInstance();
		$subject = $dictionary->Translate("PIRATES_FRIENDS_NO_MORE");
		$textBody = $dictionary->Translate("PIRATES_FRIENDS_NO_MORE_BODY");
		$message = Message::CreateNewMessage($pirates->GetId(), $player->GetId(), $subject, $textBody);
	}
	
	/**
	 * Plays for PC player if possible.
	 */
	public function PlayPCPlayer()
	{
		$pcPlayer = Player::LoadPCPlayerFromDB();
		
		$ships = $pcPlayer->GetShipsForced();
		
		foreach ($ships as $ship)
		{
			$this->PlayNPCShip($pcPlayer, $ship);
		}	
		
		$this->ManageNPCShips($pcPlayer);	
		
		$actionLogger = new ActionLogger($pcPlayer);
		$actionLogger->ClearLogForPlayer();
	}
	
	private function ManageNPCShips(Player $npcPlayer)
	{
		echo "<h3>Spravuji lode NPC...</h3>";
		$cargoCount = 0;
		$ships = $npcPlayer->GetShipsForced();
		$lastCargo = NULL;
		foreach ($ships as $ship)
		{
			if ($ship->GetModel()->GetShipType()->GetId() == GameConstants::CARGO_SHIP_TYPE_ID)
			{
				$lastCargo = $ship;
				$cargoCount++;
			}
		}
		
		printf("mam %d lodi<br/>", $cargoCount);
		$licenceCount = $npcPlayer->GetLicenceCount();
		printf("mam %d licenci<br/>", $licenceCount);
		
		while (true)
		{
			// upgrade ship
			if ($lastCargo != NULL)
			{
				$completed = $this->ManageNPCCargoShip(Player::LoadPCPlayerFromDB(), $lastCargo);
				
				if (!$completed)
				{
					printf("dosly prachy 1<br/>");
					return; // out of money
				}
			}
			
			// buy licences
			while ($licenceCount <= $cargoCount)
			{
				printf("kupuju licenci<br/>");
				if (Responder::BuyLicence(Player::LoadPCPlayerFromDB()) != 1)
				{
					printf("dosly prachy 2<br/>");
					return;
				}
				$licenceCount++;
			}			
			
			printf("kupuju lod<br/>");
			// buy ship
			$lastCargo = Responder::BuyShip(self::NPC_CARGO_SHIP_MODEL_ID, 'NPC_CARGO_' . mt_rand(0, mt_getrandmax()), Player::LoadPCPlayerFromDB(), $this->solarSystem);
			if ($lastCargo == NULL)
			{
				printf("dosly prachy 3<br/>");
				return;
			}			
			$cargoCount++;
		}
	}
	
	private function ManageNPCCargoShip(Player $npcPlayer, Ship $cargoShip)
	{
		$guards = Utils::GetAssignedGuardians($cargoShip, $npcPlayer);
		$guardCount = count($guards);
		printf("mam %d gardu<br/>", $guardCount);
		
		while ($guardCount < GameConstants::CARGO_SHIP_TYPE_MAX_LEVEL)
		{
			printf("kupuju garda<br/>");
			$newGuard = Responder::BuyShip(self::NPC_GUARD_SHIP_MODEL_ID, $cargoShip->GetName(). '_GUARD_'. $guardCount, Player::LoadPCPlayerFromDB(), $this->solarSystem);
			if ($newGuard == NULL)
			{
				return false;
			}
			
			$guardAC = new ShipActionController($npcPlayer, $newGuard);
			$guardAC->AssignGuardTask($cargoShip);
			$guardCount++;
		}
		
		$level = $cargoShip->GetLevel();
		while ($level < GameConstants::CARGO_SHIP_TYPE_MAX_LEVEL)
		{
			printf("kupuju level<br/>");
			$res = Responder::BuyShipLevel($cargoShip->GetId(), Player::LoadPCPlayerFromDB());
			echo $res . "<br/>";
			if (!($res == 1))
			{
				return false;
			}
			$level++;
		}
		
		printf("upgrade complete<br/>");
		return true;
	}
	
	/**
	 * Plays for PC player ship if possible.
	 * @param pcPlayer NPC player
	 * @param ship ship to play with
	 */
	private function PlayNPCShip(Player $pcPlayer, Ship $ship)
	{
		$shipController = new ShipActionController($pcPlayer, $ship, $this->solarSystem);
		$shipController->UpdateState();
		
		if ($shipController->IsBlocked())
		{
			return;
		}
		
		$dockedPlanet = $shipController->GetDockedPlanet();
		
		if ($dockedPlanet == NULL)
		{
			$dockedPlanet = $this->GetRandomPlanet();
			$shipController->DockToPlanet($dockedPlanet);
		}
		
		$deliveries = $dockedPlanet->GetDeliveriesForced();
		$stuffLoaded = 0;
		
		foreach ($deliveries as $delivery)
		{
			//if (!$shipController->CanBeDelivered($delivery))
			//{
			//	continue;
			//}
			
			//echo "beru ". $delivery->GetId() ."<br/>";
			
			if ($delivery->IsHidden())
			{
				continue;
			}
			
			$delivery->WriteNewLocationId(-($ship->GetId()));
			$ship->InvalidateDeliveries();
			$stuffLoaded++;
			
			if ($stuffLoaded >= 2)
			{
				break;
			}
		}
		
		$newPlanet = $this->GetRandomPlanet($dockedPlanet);
		
		$timeForTravel = FlightController::GetPlanetDistance($dockedPlanet, $newPlanet) / $ship->GetModel()->GetSpeed() * rand(1, 3);
		
		$shipController->DockToPlanet($newPlanet);
		$shipController->UnloadAll(false);		
		$shipController->BlockShip(GameTimer::EarthDaysToRealTime($timeForTravel / 3600.0 / 24 * GameConstants::NPC_BLOCK_HANDICAP));
	}
	
	/**
	 * Generates random deliveries from specified planet.
	 * @param planet start planet of delivery
	 * @param count number of deliveries to generate 
	 */
	private function GenerateDeliveriesForPlanet(Planet $planet, $count)
	{
		if ($planet == NULL)
		{
			throw new Exception("Planet must not be null.");	
		}
		
		if ($count < 0)
		{
			throw new Exception("Delivery count must not be negative.");
		}
		
		for ($i = 0; $i < $count; $i++)
		{
			$this->GenerateRandomDelivery($planet);
		}
		
		$planet->LoadDeliveriesFromDB();
	}
	
	private function GenerateRandomDelivery(Planet $planet)
	{
		$amount = rand(GameConstants::MIN_CARGO_AMOUNT, GameConstants::MAX_CARGO_AMOUNT);
		if (Utils::GetRandomProbability() <= GameConstants::DELIVERY_LOCAL_RATIO)
		{
			$targetPlanet = $this->GetRandomPlanet($planet, true);
		}
		else
		{
			$targetPlanet = $this->GetRandomPlanet($planet, false);
		}
		
		$goodType = $this->GetRandomGood();
		//echo "amount  = " . $amount . "<br/>";
		//echo "distance  = " . FlightController::GetPlanetDistance($planet, $targetPlanet) . "<br/>";
		//echo "amount  = " . $goodType->GetBasicRevenue() . "<br/>";
		$price = floor($goodType->GetBasicRevenue() * (FlightController::GetPlanetDistance($planet, $targetPlanet) / 1000000000) * $amount) ;
		$delivery = Delivery::CreateNewDelivery($planet, $targetPlanet, $goodType, $amount, $price);
		
		if ($delivery == NULL)
		{
			throw new Exception("Couldn't create delivery for planet with id = ".$planet->GetId());
		}
			
		return $delivery;
	}
	
	/**
	 * Returns random planet from solar system, but excludes specified planet.
	 * @param excludedPlanet planet to be excluded
	 */
	private function GetRandomPlanet(Planet $excludedPlanet = NULL, $currentSystemOnly = false)
	{
		if ($currentSystemOnly)
		{
			return $this->GetRandomPlanetFromSystem($excludedPlanet, $this->solarSystem);
		}
		else
		{
			do
			{
				$planet = Planet::LoadFromDBRandom();
			}
			while ($excludedPlanet != NULL && $planet->GetId() == $excludedPlanet->GetId());
			
			return $planet;
		}
	}
	
	private function GetRandomPlanetFromSystem(Planet $excludedPlanet = NULL, SolarSystem $system)
	{
		$planets = $system->GetPlanets();
		
		if (count($planets) <= 1 && $excludedPlanet != NULL)
		{
			return NULL;
		}
		
		if (count($planets) <= 0)
		{
			return NULL;
		}
		
		while (true)
		{
			$planet = $planets[array_rand($planets)];
			
			if ($planet != $excludedPlanet)
			{
				return $planet;
			}
		}
	}
	
	/**
	 * Returns random good from goods. Loads goods from DB
	 * if needed.
	 */
	private function GetRandomGood()
	{
		if ($this->goods == NULL)
		{
			$this->goods = GoodType::LoadAllFromDB();	
		}
		
		return $this->goods[array_rand($this->goods)];
	}
}

?>