<?php

require_once("Model/Player.inc");
require_once("Model/Ship.inc");
require_once("Model/ShipModel.inc");
require_once("Model/SolarSystem.inc");
require_once("Model/Planet.inc");
require_once("Control/GameTimer.inc");
require_once("Control/FlightController.inc");
require_once("Control/GameConstants.inc");
require_once("Control/ActionLogger.inc");

/**
 * Class for easy ship manipulation.
 * @author Petr Kellnhofer (A07278)
 *
 */
class ShipActionController
{
	/**
	 * Action string for flying ship.
	 */
	const FLYING_ACTION_NAME = "flying"; // where to;[planet/wormhole];last update time();
	
	const THROUGH_WORMHOLE_ACTION_NAME = "wormthrough";
	/**
	 * Action string for blocked ship.
	 */
	const BLOCKED_ACTION_NAME = "blocked"; // up to when;
	
	/**
	 * Action string for waiting ship.
	 */
	const WAITING_ACTION_NAME = "waiting"; //

	const LOAD_ACTION_NAME = "load";
	
	const UNLOAD_ACTION_NAME = "unload";
	
	const GUARD_ACTION_NAME = "guard"; // who
	
	/**
	 * Owner of ship.
	 */
	private $player;
	
	/**
	 * Controlled ship.
	 */
	private $ship;
	
	public function GetShip()
	{
		return $this->ship;
	}
	
	/**
	 * Solar system in which is ship present.
	 */
	private $solarSystem;
	
	public function GetSolarSystem()
	{
		return $this->solarSystem;
	}
	
	/**
	 * Creates new controller for ship.
	 * @param $player Owner of ship.
	 * @param $ship Controlled ship.
	 * @param $solarSystem Solar system in which is ship present.
	 */
	public function __construct(Player $player, Ship $ship, SolarSystem $solarSystem = NULL)
	{
		if ($player->GetId() != $ship->GetPlayerID())
		{
			throw new Exception("Player is not owner of the ship.");
		}
		
		$this->player = $player;
		$this->ship = $ship;
		
		if ($solarSystem == NULL || $ship->GetSolarSystemId() != $solarSystem->GetId())
		{
			$this->solarSystem = SolarSystem::LoadFromDB($ship->GetSolarSystemId());
		}
		else
		{
			$this->solarSystem = $solarSystem;
		}		
	}
	
	/**
	 * Returns true if ship is blocked.
	 */
	public function IsBlocked()
	{
		$tokens = explode(";", $this->ship->GetPendingAction());
		return !strcmp($tokens[0], ShipActionController::BLOCKED_ACTION_NAME);
	}
	
	/**
	 * Returns time when ship will be unblocked or false if it is not blocked.
	 */
	public function GetBlockEndTime()
	{
		if (!$this->IsBlocked())
		{
			return false;			
		}
		
		$tokens = explode(";", $this->ship->GetPendingAction());
		return $tokens[1];
	}
	
	/**
	 * Returns true if ship is waiting.
	 */
	public function IsWaiting()
	{
		$tokens = explode(";", $this->ship->GetPendingAction());
		
		if (count($tokens) == 0)
		{
			return true;
		}
		
		if (strlen(trim($tokens[0])) == 0)
		{
			return true;	
		}
		
		if (!strcmp($tokens[0], ShipActionController::WAITING_ACTION_NAME))
		{
			return true;
		}		
	}
	
	/**
	 * Returns true if ship is flying.
	 */
	public function IsFlying()
	{
		$tokens = explode(";", $this->ship->GetPendingAction());
		return !strcmp($tokens[0], ShipActionController::FLYING_ACTION_NAME);
	}
	
	public function IsFlyingToPlanet()
	{
		$tokens = explode(";", $this->ship->GetPendingAction());
		if (strcmp($tokens[0], ShipActionController::FLYING_ACTION_NAME))
		{
			return false;
		}
		
		return !strcmp($tokens[1], "planet");
	}
	
	public function IsFlyingToWormHole()
	{
		$tokens = explode(";", $this->ship->GetPendingAction());
		if (strcmp($tokens[0], ShipActionController::FLYING_ACTION_NAME))
		{
			return false;
		}
		
		return !strcmp($tokens[1], "wormhole");
	}
			
	public function GetTargetPlanet()
	{
		if (!$this->IsFlyingToPlanet())
		{
			return NULL;
		}
		
		$tokens = explode(";", $this->ship->GetPendingAction());		
		$planets = (array)$this->solarSystem->GetPlanets();
		
		return $planets[$tokens[2]];
	}
	
	public function GetTargetObject()
	{
		if (!$this->IsFlying())
		{
			return NULL;
		}
		
		$tokens = explode(";", $this->ship->GetPendingAction());
		
		if (!strcmp($tokens[1], "planet"))
		{
			$planets = (array)$this->solarSystem->GetPlanets();
			return $planets[$tokens[2]];
		}
		else
		{
			return WormHole::LoadFromDB($tokens[2]);
		}		
	}
	
	public function IsGuarding()
	{
		$tokens = explode(";", $this->ship->GetPendingAction());
		return !strcmp($tokens[0], ShipActionController::GUARD_ACTION_NAME);
	}
	
	public function GetGuardedShipId()
	{
		if (!$this->IsGuarding())
		{
			return -1;
		}
		
		$tokens = explode(";", $this->ship->GetPendingAction());
		return $tokens[1];
	}
	
	/**
	 * Returns planet on which the ship is docked or NULL if ship
	 * is not docked.
	 * @return planet where ship is docked or NULL
	 */
	public function GetDockedPlanet() 
	{
		if ($this->ship->GetDockedPlanetId() <= 0)
		{
			return NULL;
		}
		
		$planets = (array)$this->solarSystem->GetPlanets();
		
		if (!array_key_exists($this->ship->GetDockedPlanetId(), $planets))
		{
			return NULL;
		}	
		
		return $planets[$this->ship->GetDockedPlanetId()];
	}
	
	/**
	 * Docks ship to specified planet so it moves with planet.
	 * @param $planet planet to dock on
	 */
	public function DockToPlanet(Planet $planet)
	{
		$this->ship->WriteNewDockedPlanetId($planet->GetId());
		$this->ship->WriteNewPosition($planet->GetWorldX(), $planet->GetWorldY());
				
		if (Utils::HasPlayersAllianceDockOnPlanet($this->player, $planet))
		{
			$price = 0;
		}
		else
		{
			$price = GameConstants::DOCK_PRICE;
			$this->player->WriteNewMoney($this->player->GetMoney() - $price);
		}
		
		$logger = new ActionLogger($this->player);
		$logger->LogShipDocked($this->ship, $planet, $price);
	}
	
	/**
	 * Undocks ship so it can move freely.
	 */
	public function Undock()
	{
		$this->ship->WriteNewDockedPlanetId(0);
	}
	
	public function TravelThroughWormHole(WormHole $wormhole)
	{
		$systemIds = (array)$wormhole->GetSystemIds();
		if ($systemIds[0] != $this->ship->GetSolarSystemId() && $systemIds[1] != $this->ship->GetSolarSystemId())
		{
			return false;
		}
		
		$opositeSystemIndex = $systemIds[0] != $this->ship->GetSolarSystemId() ? 0 : 1;
		
		$this->Undock();
		$this->ship->SetSolarSystemId($systemIds[$opositeSystemIndex]);
		
		$newPos = $wormhole->GetPositions();
		$newPos = $newPos[$opositeSystemIndex];
		$this->ship->WriteNewPosition($newPos[0], $newPos[1]);
		
		$this->solarSystem = SolarSystem::LoadFromDB($this->ship->GetSolarSystemId());
		
		$taxes = Utils::CountShipDeliveriesPrice($this->ship) * GameConstants::WORMHOLE_TAXES_FROM_CARGO_RATIO;
		$this->player->WriteNewMoney($this->player->GetMoney() - $taxes);
		
		$logger = new ActionLogger($this->player);
		$logger->LogShipFlownThroughWormHole($this->ship, $this->solarSystem->GetName(), $taxes);
	}
	
	public function CanBeDelivered(Delivery $delivery)
	{
		if ($delivery->GetLocationID() < 0)
		{
			//echo "bad id";
			return false;
		}
		
		$compatibleGoods = (array)$this->ship->GetModel()->GetShipType()->GetCompatibleGoodClassesForced();
		if (!array_key_exists($delivery->GetGoodType()->GetGoodClass()->GetId(), $compatibleGoods))
		{
			//echo "bad ship type";
			return false;
		}
		
		$oldDeliveries = (array)$this->ship->GetDeliveriesForced();
		
		if (count($oldDeliveries) > 0)
		{
			$oldDelivery = array_pop($oldDeliveries);
		}
		
		if ($oldDelivery != NULL && $oldDelivery->GetTargetPlanetID() != $delivery->GetTargetPlanetID())
		{
//			if (!strcmp($this->ship->GetName(), "First ship"))
//			{
//				echo "Jede to na: ".Planet::LoadFromDB($delivery->GetTargetPlanetID())->GetName();
//			}
//			
			//echo "old cargo goes elsewhere";	
			return false;
		}
		
		$freeSpace = $this->GetFreeCargoCapacity();
		
		if ($freeSpace < $delivery->GetAmount())
		{
			//echo "lack of space";
			return false;
		}
		
		// only cargo ships can deliver between different systems
		$startPlanet = Planet::LoadFromDB($delivery->GetLocationID());
		$targetPlanet = Planet::LoadFromDB($delivery->GetTargetPlanetID());
		if (($startPlanet->GetSolarSystemId() != $targetPlanet->GetSolarSystemId())
			&& $this->ship->GetModel()->GetShipType()->GetId() != GameConstants::CARGO_SHIP_TYPE_ID)
		{
			//echo "mezikont";
			return false;		
		}
		
		return true;
	}
	
	// TODO: Not finished.
	/**
	 * Updates state of ship.
	 */
	public function UpdateState()
	{
		$actions = explode("\n", $this->ship->GetPendingAction());
		$newActions = array();
		
		$free = true;
		
		foreach ($actions as $action)
		{
			$finished = false;
			
			if ($free)
			{
				$action = $this->ProcessActionLine($action);
				$finished = (strlen($action) == 0);
			}
			
			if ($finished)
			{
				continue;
			}
			else
			{
				$free = false;
			}
			
			array_push($newActions, $action);
		}
		
		if (count($newActions) == 0) // WAITING
		{
			array_push($newActions, sprintf("%s", ShipActionController::WAITING_ACTION_NAME));
		}
		
		$actionList = "";
		for ($i = 0; $i < count($newActions); $i++)
		{		
			$actionList .= $newActions[$i];
			if ($i != (count($newActions) - 1))
			{
				$actionList .= "\n";
			}
		}
		
		$this->ship->PerformNewAction($actionList);
		
		$dockPlanet = $this->GetDockedPlanet();
		if ($dockPlanet != NULL)
		{
			$this->ship->WriteNewPosition($dockPlanet->GetWorldX(), $dockPlanet->GetWorldY());
		}
	}
	
	/**
	 * 
	 * @param $action
	 * @return updated string or empty string for completed action
	 */
	private function ProcessActionLine($action)
	{
		$now = time();
		$day = GameTimer::RealTimeToEarthDays($now - $this->ship->GetLastUpdateTime());
		$tokens = explode(";", $action);
				
		// unblock
		if (!strcmp($tokens[0], ShipActionController::BLOCKED_ACTION_NAME))
		{
			if ($tokens[1] <= time())
			{
				//echo "poustim ho";
				return "";
			}
			else 
			{
				return $action;
			}
		} 
		else if(!strcmp($tokens[0], ShipActionController::FLYING_ACTION_NAME))
		{
			if ($tokens[3] < 0)
			{
				return sprintf("%s;%s;%d;%d", ShipActionController::FLYING_ACTION_NAME, $tokens[1], $tokens[2], $now);
			}
			
			$dockedPlanet = $this->GetDockedPlanet();
			
			if (!strcasecmp($tokens[1], "planet"))
			{
				$planets = (array)$this->solarSystem->GetPlanets();
				$target = $planets[$tokens[2]];
				
				if ($dockedPlanet != NULL && $dockedPlanet->GetId() == $target->GetId())
				{
					return "";
				}
			}
			else if (!strcasecmp($tokens[1], "wormhole"))
			{
				$wormholes = (array)$this->solarSystem->GetWormHolesForced();
				$target = $wormholes[$tokens[2]];
			}
			
			if ($target == NULL)
			{
				return "";
			}
			
			if ($dockedPlanet != NULL)
			{
				$this->Undock();
			}
			
			$flighter = new FlightController($this);
		
			$elapsedDays = GameTimer::RealTimeToEarthDays(microtime(true) - $tokens[3]);
			$elapsedSecs = $elapsedDays * GameTimer::SECONDS_IN_DAY;
			$startDay = GameTimer::GetDaysElapsed(true) - $elapsedDays;
						
			$flightRes = (array)$flighter->FlyToTarget($target, $startDay, $elapsedSecs);
			$dirChange = $tokens[4] + $flightRes[3];
			
			if ($flightRes[0])
			{
				$priceForDirChange = $dirChange * GameConstants::PRICE_FOR_DIRECTION_CHANGE;
				$this->player->WriteNewMoney($this->player->GetMoney() - $priceForDirChange);
				
				$logger = new ActionLogger($this->player);
				$logger->LogShipFlightPayed($this->ship, $priceForDirChange);
				
				if (!strcasecmp($tokens[1], "planet"))
				{
					$this->DockToPlanet($target);
				}
				return "";
			}
			else
			{
				$path = (array)$flightRes[1];
				$newPos = array_pop($path);
				$this->ship->WriteNewPosition($newPos[0], $newPos[1]);
			}
			
			return sprintf("%s;%s;%d;%d;%d", ShipActionController::FLYING_ACTION_NAME, $tokens[1], $tokens[2], $now, $dirChange);
		} 
		else if(!strcmp($tokens[0], ShipActionController::LOAD_ACTION_NAME))
		{
			$delivery = Delivery::LoadFromDB($tokens[1]);
			if ($delivery != NULL)
			{
				$this->Load($delivery);
			}
			return "";
		} 
		else if(!strcmp($tokens[0], ShipActionController::UNLOAD_ACTION_NAME))
		{
			$deliveryCount = count((array)$this->ship->GetDeliveriesForced());
			$this->Unload(true);
			return "";			
		} 
		else if(!strcmp($tokens[0], ShipActionController::GUARD_ACTION_NAME))
		{
			return $action;			
		} 
		else if(!strcmp($tokens[0], ShipActionController::THROUGH_WORMHOLE_ACTION_NAME))
		{
			$wormHole = WormHole::LoadFromDB($tokens[1]);
			if ($wormHole != NULL)
			{
				$this->TravelThroughWormHole($wormHole);	
			}
			return "";		
		}
		
		else // WAITING
		{
			return "";
		}
	}
	
	/**
	 * Block ships for period of time.
	 * @param $numberOfEarthDays number of Earth days to block for
	 */
	public function BlockShip($numberOfEarthDays)
	{
		if ($numberOfEarthDays < 0)
		{
			throw new Exception("Number of days for ship blocking must be >= 0.");
		}	
		
		$action = sprintf("%s;%d", ShipActionController::BLOCKED_ACTION_NAME, time() + GameTimer::EarthDaysToRealTime($numberOfEarthDays));
		$this->ship->PerformNewAction($action);
	}
	
	/**
	 * Tries to load specified delivery. Removes it from planet.
	 * @param Delivery $delivery delivery to load
	 * @return true on success
	 */
	public function Load(Delivery $delivery)
	{
		if (!$this->CanBeDelivered($delivery))
		{
			throw new Exception("Delivery can't be loaded into this ship type!");
		}
		
		$planets = (array)$this->solarSystem->GetPlanets();
		$planet = $planets[$delivery->GetLocationID()];
		$planet->InvalidateDeliveries();
		
		$delivery->WriteNewLocationId(-($this->ship->GetId()));
		
		$deliveries = (array)$this->ship->GetDeliveries();
		
		if ($deliveries != NULL)
		{
			$deliveries[$delivery->GetId()] = $delivery;
		}
		
		$this->ship->InvalidateDeliveries();
		
		$logger = new ActionLogger($this->player);
		$logger->LogDeliveryLoaded($this->ship, $planet, $delivery);
		
		return true;
	}
	
	/**
	 * Gets amount of free capacity on ship.
	 * @return amount of free capacity on ship
	 */
	public function GetFreeCargoCapacity()
	{
		return $this->ship->GetFinalCapacity() - $this->GetCargoAmount();
	}
	
	/**
	 * Gets total amount of cargo on ship.
	 * @return total amount of cargo on ship
	 */
	public function GetCargoAmount()
	{
		$deliveries = (array)$this->ship->GetDeliveriesForced();
		
		$totalAmount = 0;
		
		foreach ($deliveries as $delivery)
		{
			$totalAmount += $delivery->GetAmount();
		}
		
		return $totalAmount;
	}
	
	/**
	 * Unloads only cargo whose target planet matches to
	 * currently docked planet. 
	 */
	public function UnloadMatching($doLog = true)
	{
		return $this->Unload(true, $doLog);
	}
	
	/**
	 * Unloads all cargo.
	 */
	public function UnloadAll($doLog = true)
	{
		return $this->Unload(false, $doLog);
	}
	
	/**
	 * Unloads cargo. Adds money for delivery to player's account.
	 * @param $matchingOnly if true only matching target / docked planet are unloaded
	 */
	private function Unload($matchingOnly, $doLog = true)
	{
		if ($doLog)
		{
			$logger = new ActionLogger($this->player);
		}
		$dockedPlanet = $this->GetDockedPlanet();
		
		if ($dockedPlanet == NULL)
		{
			return;
		}
		
		$deliveries = (array)$this->ship->GetDeliveriesForced();
		$planets = (array)$this->solarSystem->GetPlanets();
		
		$cashBonus = 0;
		
		$deliveredCount = 0;
		
		foreach ($deliveries as $delivery)
		{
			if ($matchingOnly && $delivery->GetTargetPlanetID() != $dockedPlanet->GetId())
			{
				continue;
			}
			
			$cashBonus += $delivery->GetPrice();
			$deliveredCount++;
			
			if ($doLog)
			{
				$logger->LogDeliveryUnloaded($this->ship, Planet::LoadFromDB($delivery->GetTargetPlanetID()), $delivery);
			}
			$delivery->RemoveFromDB();
		}
		
		$this->ship->InvalidateDeliveries();
		
		if ($cashBonus != 0)
		{
			$this->player->WriteNewMoney($this->player->GetMoney() + $cashBonus);
			$this->ship->WriteDeliveryDoneCount($this->ship->GetDeliveryDoneCount() + $deliveredCount);
		}
	}
	
	public function EnqueDeliverDelivery($delivery)
	{
		if ($delivery == NULL)
		{
			throw new Exception("Delivery must not be NULL.");
		}
		
		if (!$this->CanBeDelivered($delivery))
		{
			throw new Exception("Delivery can't be loaded into this ship type!");
		}
				
		if ($delivery->IsHidden())
		{
			throw new Exception("Delivery is hidden and can't be enqued'!");
		}
		
		$startId = $delivery->GetLocationID();
		
		if ($startId < 0)
		{
			throw new Exception("Location of delivery must be planet.");
		}
		
		$startPlanet = Planet::LoadFromDB($delivery->GetLocationID());
				
		$this->EnqueFlyToPlanet($startPlanet);
		$this->EnqueLoad($delivery);
		
		if ($this->ship->GetModel()->GetShipType()->GetId() != GameConstants::CARGO_SHIP_TYPE_ID)
		{
			$targetPlanet = Planet::LoadFromDB($delivery->GetTargetPlanetID());
			$this->EnqueFlyToPlanet($targetPlanet);
			$this->EnqueUnload($delivery);
		}		
		
		return true;
	}
	
	public function EnqueFlyToPlanet(Planet $planet)
	{
		if ($this->IsWaiting())
		{
			$actionList = "";
			$lastCalc = time();
		}
		else
		{
			$actionList = $this->ship->GetPendingAction() . "\n";
			$lastCalc = -1;
		}
		
		$actionList .= ShipActionController::FLYING_ACTION_NAME . ";planet;" . $planet->GetId() . ";" . $lastCalc . ";0";
		return $this->ship->PerformNewAction($actionList);
	}
	
	public function EnqueFlyToWormhole(WormHole $wormHole)
	{
		if ($this->IsWaiting())
		{
			$actionList = "";
			$lastCalc = time();
		}
		else
		{
			$actionList = $this->ship->GetPendingAction() . "\n";
			$lastCalc = -1;
		}
		
		$actionList .= ShipActionController::FLYING_ACTION_NAME . ";wormhole;" . $wormHole->GetId() . ";" . $lastCalc . ";0";
		return $this->ship->PerformNewAction($actionList);
	}
	
	public function EnqueTravelThroughWormhole(WormHole $wormHole)
	{
		if ($this->IsWaiting())
		{
			return $this->TravelThroughWormHole($wormHole);
		}
		else
		{
			$actionList = $this->ship->GetPendingAction() . "\n";
		}
		
		$actionList .= ShipActionController::THROUGH_WORMHOLE_ACTION_NAME . ";" . $wormHole->GetId();
		return $this->ship->PerformNewAction($actionList);
	}
	
	public function EnqueLoad(Delivery $delivery)
	{
		$delivery->SetHidden(true);
		if ($this->IsWaiting())
		{
			$this->Load($delivery);
		}
		else
		{
			$actionList = $this->ship->GetPendingAction() . "\n";
		}
		
		$actionList .= ShipActionController::LOAD_ACTION_NAME . ";" . $delivery->GetId();
		return $this->ship->PerformNewAction($actionList);
	}
	
	public function EnqueUnload(Delivery $delivery)
	{
		if ($this->IsWaiting())
		{
			$this->Unload(true);
		}
		else
		{
			$actionList = $this->ship->GetPendingAction() . "\n";
		}
		
		$actionList .= ShipActionController::UNLOAD_ACTION_NAME . ";" . $delivery->GetId();
		return $this->ship->PerformNewAction($actionList);
	}
	
	public function AssignGuardTask(Ship $cargoShip = NULL)
	{
		if ($cargoShip == NULL && $this->GetGuardedShipId() < 0)
		{
			return true;
		}
		
		if ($this->ship->GetModel()->GetShipType()->GetId() != GameConstants::GUARD_SHIP_TYPE_ID)
		{
			return false;
		}
		
		if (count($this->ship->GetDeliveriesForced()) > 0)
		{
			return false;
		}
		
		if ($cargoShip == NULL)
		{
			$actionList = self::WAITING_ACTION_NAME;
		}
		else
		{
			if ($cargoShip->GetModel()->GetShipType()->GetId() != GameConstants::CARGO_SHIP_TYPE_ID)
			{
				return false;
			}
			
			$actionList = self::GUARD_ACTION_NAME . ";" . $cargoShip->GetId();
		}
		
		return $this->ship->PerformNewAction($actionList);
	}
}

?>