<?php

require_once("Model/PlanetTrajectory.inc");
require_once("Model/Delivery.inc");
require_once("Model/Base.inc");
require_once("Control/DBManager.inc");
require_once("Control/DBConstants.inc");
require_once("Control/GameTimer.inc");
require_once("Control/INavigable.inc");

/**
 * Planet of solar system.
 * Image of DB entity "planety".
 */
class Planet implements INavigable
{
	/**
	 * ID of planet as in DB.
	 */
	private $id;

	/**
	 * Gets ID of planet as in DB.
	 */
	public function GetId()
	{
		return $this->id;
	}
	
	private $solarSystemId;

	/**
	 * Gets ID of planet as in DB.
	 */
	public function GetSolarSystemId()
	{
		return $this->solarSystemId;
	}

	/**
	 * Name of planet.
	 */
	private $name;
	 
	/**
	 * Gets name of planet.
	 */
	public function GetName()
	{
		return $this->name;
	}

	/**
	 * Current ingame world X coordinate of planet.
	 */
	private $worldX;
	 
	/**
	 * Gets current ingame world X coordinate of planet.
	 */
	public function GetWorldX()
	{
		return $this->worldX;
	}
	 
	/**
	 * Current ingame world Y coordinate of planet.
	 */
	private $worldY;
	 
	/**
	 * Gets current ingame world Y coordinate of planet.
	 */
	public function GetWorldY()
	{
		return $this->worldY;
	}

	/**
	 * URL to image of this planet.
	 */
	private $imageURL;

	/**
	 * Gets URL to image of this planet.
	 */
	public function GetImageURL()
	{
		return $this->imageURL;
	}
	
	/**
	 * Number of current day in planet's year.
	 */
	private $currentPlanetDay;
	
	/**
	 * Gets number of current day in planet's year.
	 */
	public function GetCurrentPlanetDay()
	{
		return $this->currentPlanetDay;
	}
	
	private $dayOffset;

	/**
	 * Length of year on this planet in number of Earth days.
	 */
	private $yearLengthInDays;

	/**
	 * Gets length of year on this planet in number of Earth days.
	 */
	public function GetYearLengthInDays()
	{
		return $this->yearLengthInDays;
	}

	/**
	 * Array of planets deliveries with IDs as keys.
	 * May be null if not loaded yet - it is loaded separately
	 * to save performance when no deliveries are needed - i.e. when just
	 * retrieving planet position.
	 */
	private $deliveries;

	/**
	 * Gets array of planets deliveries with IDs as keys.
	 * May return null if not loaded yet - it is loaded separately
	 * to save performance when no deliveries are needed - i.e. when just
	 * retrieving planet position.
	 */
	public function GetDeliveries()
	{
		return $this->deliveries;
	}
	
	/**
	 * Gets deliveries for planet.
	 * Reads from DB if necessary.
	 * @return deliveries for planet
	 */
	public function GetDeliveriesForced()
	{
		if ($this->deliveries == NULL)
		{
			$this->LoadDeliveriesFromDB();
		}
		
		return $this->deliveries;
	}
	
	/**
	 * Marks deliveries as out of date.
	 */
	public function InvalidateDeliveries()
	{
		$this->deliveries = NULL;
	}
	
	private $bases;
	
	/**
	 * Gets array of planet's bases with their IDs as keys.
	 * May return null if not loaded yet - it is loaded separately
	 * to save performance when no ships are needed - i.e. when just
	 * retrieving account state.
	 */
	public function GetBases()
	{
		return $this->bases;
	}
	
	/**
	 * Gets bases for planet.
	 * Reads from DB if necessary.
	 * @return bases for player
	 */
	public function GetBasesForced()
	{
		if ($this->bases == NULL)
		{
			$this->LoadBasesFromDB();
		}
		
		return $this->bases;
	}
		
	private $trajectory;
	
	public function GetTrajectory()
	{
		return $this->trajectory;		
	}
	 
	/**
	 * Creates new Planet with specified ID and name.
	 * @param id ID of planet
	 * @param name name of planet
	 */
	private function __construct($id, $name)
	{
		$this->id = $id;
		$this->name = $name;
		$this->deliveries = NULL;
	}
	
	public static function LoadAllFromDB()
	{
		$dbManager = DBManager::GetInstance();

		$query = sprintf("SELECT ID FROM `%s`", DBConstants::PLANET_TABLE_NAME);
		$rows = $dbManager->PerformQuery($query);
		
		if (!$rows)
		{
			return NULL;
		}
		
		$planets = array();
		
		while ($row = mysql_fetch_assoc($rows))
		{
			$planet = Planet::LoadFromDB($row["ID"]);	
			$planets[$planet->GetId()] = $planet;
		}
		
		return $planets;
	}

	/**
	 * Loads planet with specified id from DB.
	 * Does not load deliveries.
	 * @param id ID of planet in DB
	 * @return loaded planet
	 */
	public static function LoadFromDB($id)
	{
		if ($id <= 0)
		{
			throw new Exception("Id must be positive integer.");
		}

		$dbManager = DBManager::GetInstance();

		// planet data
		$query = sprintf("SELECT * FROM `%s` WHERE id=%d", DBConstants::PLANET_TABLE_NAME, $id);
		$data = $dbManager->PerformQueryAssoc($query);

		if (!$data)
		{
			return NULL;
		}

		return Planet::ParseDBRow($data);
	}
	
	public static function LoadFromDBRandom()
	{
		$dbManager = DBManager::GetInstance();
		$row = $dbManager->ReadRandomRow(DBConstants::PLANET_TABLE_NAME, "ID");
		
		return Planet::ParseDBRow($row);
	}
	
	private static function ParseDBRow($row)
	{
		$planet = new Planet((int)$row["ID"], $row["nazev"]);
		$planet->imageURL = $row["obrazek"];
		$planet->solarSystemId = $row["solar_system_id"];
		$planet->dayOffset = $row["day_offset"];
		$planet->trajectory = PlanetTrajectory::LoadFromDB($row["trajectory_id"]);

		$planet->yearLengthInDays = $planet->trajectory->GetPositionCount();
		$planet->currentPlanetDay = (GameTimer::GetDaysElapsed() + $planet->dayOffset) % $planet->yearLengthInDays;

		$position = $planet->GetPositionForGameDay(GameTimer::GetDaysElapsed());
		$planet->worldX = $position[0];
		$planet->worldY = $position[1];

		return $planet;
	} 

	/**
	 * Loads deliveries for this planet from DB.
	 * Does not return anything, use GetDeliveries() to access result.
	 */
	public function LoadDeliveriesFromDB()
	{
		$dbManager = DBManager::GetInstance();

		$query = sprintf("SELECT ID FROM `%s` WHERE planeta=%d", DBConstants::DELIVERY_TABLE_NAME, $this->id);
		$rows = $dbManager->PerformQuery($query);

		if (!$rows)
		{
			return NULL;
		}

		$this->deliveries = array();

		while ($row = mysql_fetch_assoc($rows))
		{
			$delivery = Delivery::LoadFromDB($row["ID"]);
			$this->deliveries[$delivery->GetId()] = $delivery;
		}
	}
	
	/**
	 * Gets coordinates of planet's position for next day.
	 * @return length 2 array with x a y coordinates of next position
	 */
	public function GetNextPositionFromDB()
	{
		return $this->GetPositionForGameDay(GameTimer::GetDaysElapsed() + 1);
	}
	
	public function GetPositionForGameDay($gameDay, $solarSystemId = NULL)
	{
		if ($solarSystemId != NULL && $solarSystemId != $this->solarSystemId)
		{
			return NULL;
		}
		
		return $this->trajectory->GetPosition(($gameDay + $this->dayOffset) % $this->yearLengthInDays);
	}
		
	public function GetCurrentPosition($solarSystemId = NULL)
	{
		if ($solarSystemId != NULL && $solarSystemId != $this->solarSystemId)
		{
			return NULL;
		}
		
		return array(0 => $this->GetWorldX(), 1 => $this->GetWorldY());		
	}
	
	public function LoadBasesFromDB()
	{
		$dbManager = DBManager::GetInstance();

		$query = sprintf("SELECT id FROM `%s` WHERE planet_id=%d", DBConstants::BASE_TABLE_NAME, $this->id);
		$rows = $dbManager->PerformQuery($query);

		if (!$rows)
		{
			return NULL;
		}

		$this->bases = array();

		while ($row = mysql_fetch_assoc($rows))
		{
			$base = Base::LoadFromDB($row["id"]);
			$this->bases[$base->GetId()] = $base;
		}
	}
}

?>
